# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import asyncio
from typing import Optional, Iterable, TYPE_CHECKING
from pathlib import Path
import click
import pycyphal
import yakut
from . import AppDescriptor

if TYPE_CHECKING:
    import pycyphal.application  # pylint: disable=ungrouped-imports
    import pycyphal.application.file  # pylint: disable=ungrouped-imports

_logger = yakut.get_logger(__name__)


def _validate_root_directory(ctx: click.Context, param: click.Parameter, value: Iterable[str]) -> list[Path]:
    _ = param
    out: list[Path] = []
    for x in value:
        p = Path(x).resolve()
        if not p.is_dir() or not p.exists():
            raise click.UsageError(f"The specified root is not a valid directory: {x!r}", ctx=ctx)
        out.append(p)
    if not out:
        out.append(Path.cwd().resolve())  # This is the default.
    _logger.info("File server root directories: %r", list(map(str, out)))
    return out


@yakut.subcommand(aliases="fsrv")
@click.argument(
    "roots",
    metavar="PATH",
    type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=str),
    callback=_validate_root_directory,
    nargs=-1,
)
@click.option(
    "--plug-and-play",
    "-P",
    metavar="FILE",
    type=click.Path(dir_okay=False, resolve_path=True, path_type=str),
    help=f"""
Run a centralized plug-and-play (PnP) node-ID allocator alongside the file server.
The file path points to the allocation table; if missing, a new file will be created.

The PnP allocator will be tracking the status of nodes and requesting uavcan.node.GetInfo automatically.

Low-level implementation details are available in the documentation for pycyphal.application.plug_and_play
at https://pycyphal.readthedocs.io.
""",
)
@click.option(
    "--update-software",
    "-u",
    type=int,
    is_flag=False,
    flag_value=-1,
    multiple=True,
    metavar="[NODE_ID]",
    help=f"""
Check if all online nodes are running up-to-date software; request update if not.
The software version is determined by invoking uavcan.node.GetInfo for every node that is online or
became online (or restarted).

This option may take an optional value which shall be a node-ID,
in which case the specified node will be explicitly commanded to begin an update once regardless of its state
(even if an update is already in progress, but this is done only once to avoid infinite loop and only
if a newer software image is available),
while no other nodes will be updated.
This option can be given multiple times to select multiple nodes to update;
if it is given at least once without a value, all nodes will be considered candidates for an update.

When a node responds to uavcan.node.GetInfo, the root directory of the file server is scanned for software packages
that are suitable for the node.
If the node is already executing one of the available software packages or no suitable packages are found,
no further action is taken, as it is considered to be up-to-date.
Otherwise, if it is determined that the node should be updated, a standard software update request
of type uavcan.node.ExecuteCommand with COMMAND_BEGIN_SOFTWARE_UPDATE is sent.

To be considered by the server, node software packages shall reside in one of the root directories and
be named following this pattern:

\b
    NAME-HWMAJ.HWMIN-SWMAJ.SWMIN.VCS.CRC.app*
              |____|                |__|
        |__________|            |______|
    either minor or both        either CRC or both
    can be omitted if           CRC and VCS can be
    multiple hardware           omitted
    versions supported

The values are sourced from uavcan.node.GetInfo, and they are as follows:

NAME -- The name of the node; e.g., "com.zubax.telega".

HWMAJ, HWMIN -- Hardware version numbers.
The minor number or both of them can be omitted iff the package is compatible with multiple hardware revisions.

SWMAJ, SWMIN -- Software version numbers.

VCS, CRC --
The version control system (VCS) revision ID (e.g., git commit hash) and the CRC of the software package.
Both are hexadecimal numbers and both are optional: either the CRC alone or both VCS-hash and CRC can be omitted.

The fields are terminated by a literal string ".app",
which can be followed by arbitrary additional metadata (like a file extension).

Examples of compliant names:

\b
    com.zubax.telega-1.2-0.3.68620b82.28df0c432c2718cd.app.bin
    com.zubax.telega-0.3.app.zip

A node running software version X (as determined from uavcan.node.GetInfo)
is considered to require an update to Y (a local package file) if
the names are matching, the hardware version is compatible, and either condition holds:

- The software image CRC is defined for both X and Y and is different.

- The software version of Y is higher than X.

- The software version of Y is not older than X and the VCS hash is different.

- There may be additional heuristics to handle edge cases. Inspect logs or use --verbose to see details.
""",
)
@yakut.pass_purser
@yakut.asynchronous(interrupted_ok=True)
async def file_server(
    purser: yakut.Purser,
    roots: list[Path],
    plug_and_play: Optional[str],
    update_software: tuple[int, ...],
) -> None:
    """
    Run a standard Cyphal file server; optionally run a plug-and-play node-ID allocator and software updater.

    The command takes a list of root directories for the file server.
    If none are given, the current working directory will be used as the only root.
    If more than one root is given, they all will be visible via Cyphal as a single unified directory;
    the first directory takes precedence in case of conflicting entries.

    Examples:

    \b
        yakut file-server --plug-and-play=allocation_table.db --update-software
    """
    try:
        from pycyphal.application import NodeInfo
        from pycyphal.application.file import FileServer
        from pycyphal.application.node_tracker import NodeTracker, Entry
        from uavcan.node import ExecuteCommand_1_1 as ExecuteCommand
        from uavcan.node import Heartbeat_1_0 as Heartbeat
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name))

    update_all_nodes = any(x < 0 for x in update_software)
    explicit_nodes = set(filter(lambda x: x >= 0, update_software))
    _logger.info("Total update: %r; trigger explicitly: %r", update_all_nodes, sorted(explicit_nodes))
    with purser.get_node("file_server", allow_anonymous=False) as node:
        node_tracker: Optional[NodeTracker] = None  # Initialized lazily only if needed.

        def get_node_tracker() -> NodeTracker:
            nonlocal node_tracker
            if node_tracker is None:
                _logger.info("Initializing the node tracker")
                node_tracker = NodeTracker(node)
            return node_tracker

        if plug_and_play:
            _logger.info("Starting a plug-and-play allocator using file %r", plug_and_play)
            from pycyphal.application.plug_and_play import CentralizedAllocator

            alloc = CentralizedAllocator(node, plug_and_play)

            # The allocator requires integration with the node tracker, as explained in the docs.
            def register_node(node_id: int, _old_entry: Optional[Entry], entry: Optional[Entry]) -> None:
                if entry:
                    _logger.info("Node %r most recent heartbeat: %s", node_id, entry.heartbeat)
                    _logger.info("Node %r info: %s", node_id, entry.info or "<not available>")
                else:
                    _logger.info("Node %r went offline", node_id)
                unique_id = entry.info.unique_id.tobytes() if entry and entry.info else None
                alloc.register_node(node_id, unique_id)

            get_node_tracker().add_update_handler(register_node)

        fs = FileServer(node, roots)

        def check_software_update(node_id: int, _old_entry: Optional[Entry], entry: Optional[Entry]) -> None:
            if entry is None or entry.info is None:
                _logger.debug("Info for node %r is not (yet) available, cannot check software version", node_id)
                return
            heartbeat = entry.heartbeat
            assert isinstance(heartbeat, Heartbeat)
            if node_id not in explicit_nodes and not update_all_nodes:
                _logger.warning(
                    "Node %r ignored because total update mode not selected and its ID is not given explicitly "
                    "(explicit node-IDs are: %s)",
                    node_id,
                    ",".join(map(str, sorted(explicit_nodes))) or "(none given)",
                )
                return
            if (
                heartbeat.mode.value == heartbeat.mode.SOFTWARE_UPDATE
                and heartbeat.health.value < heartbeat.health.WARNING
            ) and node_id not in explicit_nodes:
                # Do not skip an update request if the health is WARNING because it indicates a problem, possibly
                # caused by a missing application. Details: https://github.com/OpenCyphal/yakut/issues/27
                _logger.warning(
                    "Node %r does not require an update because it is in the software update mode already, "
                    "we are not instructed to trigger it explicitly, "
                    "and its health is acceptable: %r",
                    node_id,
                    heartbeat,
                )
                return
            _logger.info("Checking if node %r requires a software update...", node_id)
            info = entry.info
            assert isinstance(info, NodeInfo)
            package = _locate_package(fs, info)
            if package:
                _local_root_is_irrelevant, remote_visible_path = package
                cmd_request = ExecuteCommand.Request(
                    ExecuteCommand.Request.COMMAND_BEGIN_SOFTWARE_UPDATE,
                    str(remote_visible_path),
                )
                _logger.warning("Requesting node %r to update its software: %r", node_id, cmd_request)
                cmd_client = node.make_client(ExecuteCommand, node_id)
                cmd_client.priority = pycyphal.transport.Priority.SLOW
                cmd_client.response_timeout = 5.0

                async def do_call() -> None:
                    result = await cmd_client.call(cmd_request)
                    if result is None:
                        _logger.error(
                            "Node %r did not respond to software update command %r in %.1f seconds",
                            node_id,
                            cmd_request,
                            cmd_client.response_timeout,
                        )
                        return
                    response, _ = result
                    assert isinstance(response, ExecuteCommand.Response)
                    if response.status != 0:
                        _logger.error(
                            "Node %r responded to software update command %r with error %r",
                            node_id,
                            cmd_request,
                            response.status,
                        )
                        return
                    _logger.info("Node %r confirmed software update command %r", node_id, cmd_request)
                    try:
                        explicit_nodes.remove(node_id)
                    except LookupError:
                        pass

                asyncio.create_task(do_call())
            else:
                _logger.warning("Node %r does not require a software update.", node_id)

        if update_software:
            _logger.info("Initializing the software update checker")
            # The check should be run in a separate thread because on a system with slow/busy disk IO this may cause
            # the file server to slow down significantly because the event loop would be blocked here on disk reads.
            get_node_tracker().add_update_handler(check_software_update)
        else:
            _logger.info("Software update checker is not required")

        await asyncio.sleep(1e100)


def _locate_package(
    fs: pycyphal.application.file.FileServer,
    info: pycyphal.application.NodeInfo,
) -> Optional[tuple[Path, Path]]:
    """
    If at least one locally available application file is equivalent to the already running application,
    no update will take place.
    This is to support the case where the network may contain nodes running several different versions
    of the application.
    Also, without this capability, if the lookup roots contained more than one application package for a
    given node, they would be continuously replacing one another.
    """
    app = AppDescriptor.from_node_info(info)
    result: Optional[tuple[Path, Path]] = None
    for root, tail in fs.glob(app.make_glob_expression()):
        candidate = AppDescriptor.from_file_name(str(tail.name))
        if candidate:
            if app.is_equivalent(candidate):
                return None
            if app.should_update_to(candidate):
                result = root, tail
    return result



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import re
import asyncio
import logging
from typing import TYPE_CHECKING
import pycyphal
from yakut.register import fetch_registers

if TYPE_CHECKING:
    import pycyphal.application


class SubjectResolver:
    _DISCOVERY_TIMEOUT = 3.0
    _RESPONSE_TIMEOUT = 1.0

    def __init__(self, local_node: "pycyphal.application.Node") -> None:
        from uavcan.node import Heartbeat_1
        from pycyphal.application.register import ValueProxy

        self._local_node = local_node
        self._reg_cache: dict[int, dict[str, ValueProxy] | None] = {}
        self._seen_nodes: set[int] = set()
        self._sub_heart = self._local_node.make_subscriber(Heartbeat_1)
        self._sub_heart.transport_session.transfer_id_timeout = 1e-3
        self._sub_heart.receive_in_background(
            lambda _, meta: self._seen_nodes.add(meta.source_node_id) if meta.source_node_id is not None else None
        )
        self._node_discovery_deadline = asyncio.get_running_loop().time() + SubjectResolver._DISCOVERY_TIMEOUT

    async def dtypes_by_id(self, subject_id: int) -> set[str]:
        """
        Maps the given subject-ID to the data types that are used with this subject across the network.
        Empty set means resolution failure -- either no nodes use this subject or they don't support
        the register interface.
        """
        await self._update_reg_cache()
        return _register_dtypes_by_id(
            {k: v for k, v in self._reg_cache.items() if v is not None},
            subject_id,
        )

    async def _update_reg_cache(self) -> None:
        while True:
            remaining = self._seen_nodes - set(self._reg_cache.keys())
            try:
                nid = next(iter(remaining))
            except StopIteration:
                if asyncio.get_running_loop().time() > self._node_discovery_deadline:
                    break
                await asyncio.sleep(0.1)  # Wait for new nodes to come up online, if any.
            else:
                self._reg_cache[nid] = await fetch_registers(
                    self._local_node.presentation,
                    nid,
                    predicate=lambda name: any(
                        x.match(name)
                        for x in [
                            _REGEX_REG_PUBSUB_NAME_ID,
                            _REGEX_REG_PUBSUB_NAME_TYPE,
                        ]
                    ),
                    timeout=SubjectResolver._RESPONSE_TIMEOUT,
                    priority=pycyphal.transport.Priority.HIGH,
                )

    def close(self) -> None:
        """
        It is mandatory to close the instance after use.
        """
        self._sub_heart.close()


_REGEX_REG_PUBSUB_NAME_ID = re.compile(r"uavcan\.(pub|sub)\.(.+)\.id")
_REGEX_REG_PUBSUB_NAME_TYPE = re.compile(r"uavcan\.(pub|sub)\.(.+)\.type")


def _register_dtypes_by_id(
    registers_per_node: dict[int, dict[str, "pycyphal.application.register.ValueProxy"]],
    subject_id: int,
) -> set[str]:
    from pycyphal.application.register import ValueConversionError

    result: set[str] = set()
    names: dict[int, str] = {}
    for node_id, registers in registers_per_node.items():
        for reg_name, reg_val in registers.items():
            match = _REGEX_REG_PUBSUB_NAME_ID.match(reg_name)
            if match:
                _pubsub, port_name = match.groups()
                try:
                    if int(reg_val) == subject_id:
                        names[node_id] = port_name
                except ValueConversionError:
                    _logger.warning("Register %r@%r contains an invalid port-ID value %r", reg_name, node_id, reg_val)
    _logger.debug("Names of subject %r per node: %r", subject_id, names)
    for node_id, port_name in names.items():
        for reg_name, reg_val in registers_per_node[node_id].items():
            match = _REGEX_REG_PUBSUB_NAME_TYPE.match(reg_name)
            if match and match.group(2) == port_name:
                try:
                    result.add(str(reg_val))
                except ValueConversionError:
                    _logger.warning("Register %r@%r contains an invalid data type name %r", reg_name, node_id, reg_val)
    return result


_logger = logging.getLogger(__name__)


def _unittest_register_dtypes_by_id() -> None:
    from tests.dsdl import ensure_compiled_dsdl

    ensure_compiled_dsdl()
    from pycyphal.application.register import ValueProxy, Natural16, String

    assert _register_dtypes_by_id({}, 123) == set()
    regs: dict[int, dict[str, ValueProxy]] = {
        0: {
            "uavcan.pub.aa.id": ValueProxy(Natural16([1000])),
            "uavcan.pub.aa.type": ValueProxy(String("ns.A.1.1")),
            # aa
            "uavcan.sub.bb.id": ValueProxy(Natural16([2000])),
            "uavcan.sub.bb.type": ValueProxy(String("ns.B.1.1")),
            # bb
            "uavcan.sub.typeless.id": ValueProxy(Natural16([3000])),
            # bad ID
            "uavcan.pub.bad_id.id": ValueProxy(String("not a number")),
            # bad type
            "uavcan.pub.bad_type.id": ValueProxy(Natural16([2000])),
            "uavcan.pub.bad_type.type": ValueProxy(Natural16([2000])),
        },
        1: {
            "uavcan.sub.cc.id": ValueProxy(Natural16([2000])),
            "uavcan.sub.cc.type": ValueProxy(String("ns.B.1.1")),
            "uavcan.pub.aa.id": ValueProxy(Natural16([1000])),
            "uavcan.pub.aa.type": ValueProxy(String("ns.A.2.2")),
        },
        3: {},
    }
    assert _register_dtypes_by_id(regs, 1000) == {"ns.A.1.1", "ns.A.2.2"}
    assert _register_dtypes_by_id(regs, 2000) == {"ns.B.1.1"}  # Bad type ignored.
    assert _register_dtypes_by_id(regs, 3000) == set()  # Typeless ignored.
    assert _register_dtypes_by_id(regs, 9000) == set()  # Not found



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import math
from typing import Tuple, List, Sequence, Callable, Any, Dict
import dataclasses
import logging
import textwrap
from functools import lru_cache
import click
import pycyphal
import yakut
from yakut.enum_param import EnumParam
from yakut.yaml import EvaluableLoader
from yakut.subject_specifier_processor import process_subject_specifier, SubjectResolver
from ._executor import Executor, Publication

_MIN_SEND_TIMEOUT = 0.1
"""
With a slow garbage-collected language like Python, having a smaller timeout does not make practical sense.
This may be made configurable later.
"""


@dataclasses.dataclass(frozen=True)
class ExpressionContextModule:
    """
    Defines a Python module that is imported and available for use within YAML-embedded user expressions.
    """

    module_name: str
    docs_uri: str
    alias: str = ""
    wildcard: bool = False

    def __str__(self) -> str:
        if self.wildcard:
            body = f"from {self.module_name} import *"
        elif self.alias:
            body = f"import {self.module_name} as {self.alias}"
        else:
            body = f"import {self.module_name}"
        return body.ljust(25) + f"# {self.docs_uri}"

    @staticmethod
    def load(what: Sequence[ExpressionContextModule]) -> Dict[str, Any]:
        import inspect

        out: Dict[str, Any] = {}
        for ecm in what:
            _logger.debug("Importing: %r", ecm)
            mod = __import__(ecm.module_name)
            out[ecm.module_name] = mod
            if ecm.alias:
                out[ecm.alias] = mod
            if ecm.wildcard:
                out.update({name: member for name, member in inspect.getmembers(mod) if not name.startswith("_")})

        _logger.debug("Expression context contains %d items (listed on the next line):\n%s", len(out), list(out))
        return out


_EXPRESSION_CONTEXT_MODULES = [
    ExpressionContextModule("random", "https://docs.python.org/library/random.html", wildcard=True),
    ExpressionContextModule("time", "https://docs.python.org/library/time.html", wildcard=True),
    ExpressionContextModule("math", "https://docs.python.org/library/math.html", wildcard=True),
    ExpressionContextModule("os", "https://docs.python.org/library/os.html"),
    ExpressionContextModule("pycyphal", "https://pycyphal.readthedocs.io"),
    ExpressionContextModule("numpy", "https://numpy.org", alias="np"),
    ExpressionContextModule("scipy.constants", "https://docs.scipy.org/doc/scipy/reference/constants.html"),
    ExpressionContextModule("scipy.interpolate", "https://docs.scipy.org/doc/scipy/reference/interpolate.html"),
    ExpressionContextModule("scipy.linalg", "https://docs.scipy.org/doc/scipy/reference/linalg.html"),
    ExpressionContextModule("scipy.spatial", "https://docs.scipy.org/doc/scipy/reference/spatial.html"),
]

_EXAMPLES = """
Example: publish constant messages (no embedded expressions, just regular YAML):

\b
    yakut pub uavcan.diagnostic.record '{text: "Hello world!", severity: {value: 4}}' -N3 -T0.1 -P hi
    yakut pub 33:uavcan.si.unit.angle.scalar 2.31 uavcan.diagnostic.Record 'text: "2.31 radian"'

Example: publish sinewave with frequency 1 Hz, amplitude 10 meters:

\b
    yakut pub -T 0.01 1234:uavcan.si.unit.length.scalar '!$ "sin(t * pi * 2) * 10"'

Example: as above, but control the frequency of the sinewave and its amplitude using sliders 10 and 11
of the first connected controller (use `yakut joystick` to find connected controllers and their axis mappings):

\b
    yakut pub -T 0.01 1234:uavcan.si.unit.length.Scalar '{meter: !$ "sin(t * pi * 2 * A(1,10)) * 10 * A(1,11)"}'

Example: publish 3D angular velocity setpoint, thrust setpoint, and the arming switch state:

\b
    yakut pub -T 0.1 \\
        5:uavcan.si.unit.angular_velocity.Vector3 '!$ "[A(1,0)*10, A(1,1)*10, (A(1,2)-A(1,5))*5]"' \\
        6:uavcan.si.unit.power.Scalar '!$ A(2,10)*1e3' \\
        7:uavcan.primitive.scalar.Bit '!$ T(1,5)'

Example: simulate timestamped measurement of voltage affected by white noise with standard deviation 0.25 V:

\b
    yakut pub -T 0.1 6:uavcan.si.sample.voltage.scalar \\
        '{timestamp: !$ time()*1e6, volt: !$ "A(2,10)*100+normalvariate(0,0.25)"}'
""".strip()

_HELP = f"""
Publish messages on the specified subjects.
Unless the local transport is configured in anonymous node,
the local node will also publish on standard subjects like Heartbeat and provide some standard RPC-services
like GetInfo.

The command accepts a list of space-separated pairs like:

\b
    SUBJECT_SPECIFIER  YAML_FIELDS

The first element -- SUBJECT_SPECIFIER -- defines the subject-ID and/or the data type name;
refer to the subscription command for details.

The second element -- YAML_FIELDS -- specifies the values of the message fields in YAML format
(or JSON, which is a subset of YAML).
Missing fields will be left at their default values.
For more info about the format see DSDL API docs at https://pycyphal.readthedocs.io.

The number of such pairs can be arbitrary; all defined messages will be published synchronously.
If no such pairs are specified, only the heartbeat will be published, unless the local node is anonymous.

The YAML document may embed arbitrary Python expressions that are re-evaluated immediately before publication.
Such expressions are annotated with the YAML tag `!$` (exclamation followed by dollar).
The result of such expression is substituted into the original YAML structure;
as such, the result can be of arbitrary type as long as the final YAML structure can be applied to the specified
DSDL instance.

The YAML-embedded expressions have access to the following variables:

{Executor.SYM_INDEX}: int --- index of the current publication cycle, zero initially.

{Executor.SYM_TIME}: float --- time elapsed since first message (t=n*period).

{Executor.SYM_DTYPE}: type --- message class.

{Executor.SYM_CTRL_AXIS}: (controller,axis:int)->float ---
read the normalized value of `axis` from `controller` (e.g., joystick or MIDI fader).
To see the list of available controllers and determine their channel mapping, refer to `yakut joystick`.

{Executor.SYM_CTRL_BUTTON}: (controller,button:int)->bool ---
read the state of `button` from `controller` (true while held down).

{Executor.SYM_CTRL_TOGGLE}: (controller,toggle:int)->bool ---
read the state of `toggle` from `controller`.

The following Python modules are imported and usable within embedded expressions;
refer to their API docs for usage info:

\b
{textwrap.indent(chr(0x0A).join(map(str, _EXPRESSION_CONTEXT_MODULES)), " " * 4)}

{_EXAMPLES}
"""


def _validate_message_spec(
    ctx: click.Context,
    param: click.Parameter,
    value: Tuple[str, ...],
) -> List[Tuple[str, str]]:
    if len(value) % 2 != 0:
        raise click.BadParameter(
            f"Message specifier shall have an even number of paired arguments (found {len(value)} arguments)",
            ctx=ctx,
            param=param,
        )
    return [(s, f) for s, f in (value[i : i + 2] for i in range(0, len(value), 2))]  # pylint: disable=R1721


@yakut.subcommand(help=_HELP, aliases=["pub", "p"])
@click.argument(
    "message",
    type=str,
    callback=_validate_message_spec,
    metavar="SUBJECT FIELDS [SUBJECT FIELDS]...",
    nargs=-1,
)
@click.option(
    "--period",
    "-T",
    type=float,
    default=1.0,
    show_default=True,
    metavar="SECONDS",
    help=f"""
Message publication period.
All messages are published synchronously, so the period setting applies to all specified subjects.
The send timeout equals the period as long as it is not less than {_MIN_SEND_TIMEOUT} seconds.
""",
)
@click.option(
    "--count",
    "-N",
    type=int,
    default=2**64 - 1,
    metavar="CARDINAL",
    help=f"Number of publication cycles before exiting normally. Unlimited by default.",
)
@click.option(
    "--priority",
    "-P",
    default=pycyphal.presentation.DEFAULT_PRIORITY,
    type=EnumParam(pycyphal.transport.Priority),
    help=f"Priority of published message transfers. [default: {pycyphal.presentation.DEFAULT_PRIORITY.name}]",
)
@yakut.pass_purser
@yakut.asynchronous(interrupted_ok=True)
async def publish(
    purser: yakut.Purser,
    message: Sequence[Tuple[str, str]],
    period: float,
    count: int,
    priority: pycyphal.transport.Priority,
) -> None:
    _logger.debug("period=%s, count=%s, priority=%s, message=%s", period, count, priority, message)
    assert all((isinstance(a, str) and isinstance(b, str)) for a, b in message)
    assert isinstance(period, float) and isinstance(count, int) and isinstance(priority, pycyphal.transport.Priority)
    if period < 1e-9 or not math.isfinite(period):
        raise click.BadParameter("Period shall be a positive real number of seconds")
    if count <= 0:
        _logger.warning("Nothing to do because count=%s", count)
        return
    try:
        from pycyphal.application import Node
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name))

    finalizers: list[Callable[[], None]] = []
    try:
        # Parse the field specs we were given to validate syntax.
        # Do this early before other resources are initialized.
        yaml_loader = EvaluableLoader(ExpressionContextModule.load(_EXPRESSION_CONTEXT_MODULES))
        evaluators: list[Callable[..., Any]] = []
        for _, field_spec in message:
            try:
                evaluators.append(yaml_loader.load_unevaluated(field_spec))
            except ValueError as ex:
                raise click.BadParameter(f"Invalid field spec {field_spec!r}: {ex}") from None

        # Node construction should be delayed as much as possible to avoid unnecessary interference
        # with the bus and hardware. This is why we use the factory here instead of constructing the node eagerly.
        @lru_cache(None)
        def get_node() -> Node:
            node = purser.get_node("publish", allow_anonymous=True)
            finalizers.append(node.close)
            return node

        @lru_cache(None)
        def get_subject_resolver() -> SubjectResolver:
            node = get_node()
            if node.id is None:
                raise click.ClickException(
                    f"Cannot use automatic discovery because the local node is anonymous, "
                    f"so it cannot access the introspection services on remote nodes. "
                    f"You need to either fully specify the subjects explicitly or assign a local node-ID."
                )
            sr = SubjectResolver(node)
            finalizers.append(sr.close)
            return sr

        # Resolve subject-IDs and dtypes. This may or may not require the local node.
        subject_id_dtype_pairs: list[tuple[int, Any]] = [
            await process_subject_specifier(subject_spec, get_subject_resolver) for subject_spec, _ in message
        ]

        # We delayed node construction as much as possible. It may already be constructed if subject resolver was used.
        send_timeout = max(_MIN_SEND_TIMEOUT, period)
        node = get_node()
        publications = [
            Publication(
                subject_id=sbj_id,
                dtype=dty,
                evaluator=evl,
                node=node,
                priority=priority,
                send_timeout=send_timeout,
            )
            for (sbj_id, dty), evl in zip(subject_id_dtype_pairs, evaluators)
        ]
        executor = Executor(loader=yaml_loader, publications=publications)
        finalizers.append(executor.close)

        # Everything is ready, the node can be started now. It will be stopped during finalization.
        node.start()
        _logger.info(
            "Publishing %d subjects with period %.3fs, send timeout %.3fs, count %d, priority %s",
            len(message),
            period,
            send_timeout,
            count,
            priority.name,
        )
        try:
            await executor.run(count=count, period=period)
        finally:
            _log_final_report(node.presentation)
    finally:
        pycyphal.util.broadcast(finalizers[::-1])()


def _log_final_report(presentation: pycyphal.presentation.Presentation) -> None:
    if _logger.isEnabledFor(logging.INFO):
        _logger.info("%s", presentation.transport.sample_statistics())
        for s in presentation.transport.output_sessions:
            ds = s.specifier.data_specifier
            if isinstance(ds, pycyphal.transport.MessageDataSpecifier):
                _logger.info("Subject %d: %s", ds.subject_id, s.sample_statistics())


_logger = yakut.get_logger(__name__)



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import dataclasses
from typing import TYPE_CHECKING, Any, Union, Callable, Optional, Iterable, Mapping, Sequence
import yakut
from yakut.register import unexplode_value

if TYPE_CHECKING:
    from pycyphal.application.register import Value


Unexploder = Callable[["Value"], Optional["Value"]]
"""
Returns None if the value is not coercible to the prototype.
"""

RegisterDirective = Union["Value", Unexploder]


class InvalidDirectiveError(ValueError):
    pass


SCHEMA_USER_DOC = """
If node-IDs are given explicitly, the following schemas are accepted:

\b
    [register_name]
    {register_name->register_value}

If node-IDs are not given, they shall be contained in the input:

\b
    {node_id->[register_name]}
    {node_id->{register_name->register_value}}
"""


@dataclasses.dataclass(frozen=True)
class Directive:
    registers_per_node: dict[int, dict[str, RegisterDirective]]
    """
    The directive contains either values or factories that take a prototype and return the value (or None on error).
    When a value is available, it is to be written immediately in one request.
    Otherwise, a read request will need to be executed first to discover the type of the register;
    the factory will then use that information to perform coercion (which may fail).
    """

    @staticmethod
    def load(ast: Any, node_ids: Iterable[int] | None) -> Directive:
        node_ids = list(sorted(node_ids)) if node_ids is not None else None
        if node_ids is not None:
            ast = {n: ast for n in node_ids}
            _logger.debug("Decorated: %r", ast)

        if isinstance(ast, Mapping):
            registers_per_node: dict[int, dict[str, RegisterDirective]] = {}
            for node_id_orig, node_spec in ast.items():
                try:
                    nid = int(node_id_orig)
                except ValueError:
                    raise InvalidDirectiveError(f"Not a valid node-ID: {node_id_orig}") from None
                nd = _load_node(node_spec)
                registers_per_node[nid] = nd
                _logger.debug("Loaded node directive for %d: %r", nid, nd)
            return Directive(registers_per_node=registers_per_node)

        if ast is None:
            _logger.warning("Empty directive, nothing to do")
            return Directive({})

        raise InvalidDirectiveError(f"Invalid directive: expected mapping (node_id->...), found {type(ast).__name__}")


def _load_node(ast: Any) -> dict[str, RegisterDirective]:
    from pycyphal.application.register import Value

    if isinstance(ast, Sequence) and all(isinstance(x, str) for x in ast):
        return {str(x): Value() for x in ast}

    if isinstance(ast, Mapping) and all(isinstance(x, str) for x in ast.keys()):
        return {reg_name: _load_leaf(reg_spec) for reg_name, reg_spec in ast.items()}

    if ast is None:
        return {}

    raise InvalidDirectiveError(
        f"Invalid node specifier: expected [register_name] or (register_name->register_value) or null; "
        f"found {type(ast).__name__}"
    )


def _load_leaf(exploded: Any) -> RegisterDirective:
    return unexplode_value(exploded) or (lambda proto: unexplode_value(exploded, proto))


_logger = yakut.get_logger(__name__)


def _unittest_directive() -> None:
    from pytest import raises
    from pycyphal.application.register import Value, String, Integer32, Bit

    class CV(Value):  # type: ignore
        def __eq__(self, other: object) -> bool:
            if isinstance(other, Value):
                return repr(self) == repr(other)
            return NotImplemented

    # Load full form
    dr = Directive.load(
        {
            "0": {
                "a": {"string": {"value": "z"}, "_meta_": {"ignored": "very ignored"}},
            },
            1: {
                "b": {"string": {"value": "y"}},
                "c": None,
                "d": {"empty": {}},
            },
            " 2 ": ["e", "f"],
            3: None,
        },
        node_ids=None,
    )
    assert dr.registers_per_node == {
        0: {"a": CV(string=String("z"))},
        1: {"b": CV(string=String("y")), "c": CV(), "d": CV()},
        2: {"e": CV(), "f": CV()},
        3: {},
    }

    # Load full form with explicit node-IDs
    dr = Directive.load(
        {
            "a": {"string": {"value": "z"}},
            "b": None,
        },
        node_ids=[0, 1],
    )
    assert dr.registers_per_node == {
        0: {"a": CV(string=String("z")), "b": CV()},
        1: {"a": CV(string=String("z")), "b": CV()},
    }

    # Load names only with explicit node-IDs
    dr = Directive.load(
        ["a", "b"],
        node_ids=[0, 1],
    )
    assert dr.registers_per_node == {
        0: {"a": CV(), "b": CV()},
        1: {"a": CV(), "b": CV()},
    }

    # Load simplified form, deferred callables returned.
    dr = Directive.load(
        {
            "0": {
                "a": [0, 1, 2],
                "b": 456,
            },
        },
        node_ids=None,
    )
    assert dr and len(dr.registers_per_node) == 1
    assert {"a", "b"} == dr.registers_per_node[0].keys()
    uxp = dr.registers_per_node[0]["a"]
    assert callable(uxp)
    assert CV(integer32=Integer32([0, 1, 2])) == uxp(Value(integer32=Integer32([0] * 3)))
    assert CV(bit=Bit([False, True, True])) == uxp(Value(bit=Bit([False] * 3)))
    uxp = dr.registers_per_node[0]["b"]
    assert callable(uxp)
    assert CV(integer32=Integer32([456])) == uxp(Value(integer32=Integer32([0])))

    # Errors.
    with raises(InvalidDirectiveError):
        Directive.load([], node_ids=None)
    with raises(InvalidDirectiveError):
        Directive.load("", node_ids=None)
    with raises(InvalidDirectiveError):
        Directive.load({"z": []}, node_ids=None)
    with raises(InvalidDirectiveError):
        Directive.load({"z": "q"}, node_ids=None)



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import dataclasses
from typing import Sequence, TYPE_CHECKING, Union, Optional, Callable
import pycyphal
import yakut


if TYPE_CHECKING:
    import pycyphal.application
    from uavcan.register import Access_1


@dataclasses.dataclass
class Result:
    value_per_node: dict[int, Optional["Access_1.Response"]] = dataclasses.field(default_factory=dict)
    errors: list[str] = dataclasses.field(default_factory=list)
    warnings: list[str] = dataclasses.field(default_factory=list)


async def access(
    local_node: "pycyphal.application.Node",
    progress: Callable[[str], None],
    node_ids: Sequence[int],
    *,
    reg_name: str,
    reg_val_str: str | None,
    optional_service: bool,
    optional_register: bool,
    timeout: float,
) -> Result:
    res = Result()
    for nid, item in (await _access(local_node, progress, node_ids, reg_name, reg_val_str, timeout=timeout)).items():
        _logger.debug("Register @%r: %r", nid, item)
        res.value_per_node[nid] = None  # Error state is default state
        if isinstance(item, _NoService):
            if optional_service:
                res.warnings.append(f"Service not accessible at node {nid}, ignoring as requested")
            else:
                res.errors.append(f"Service not accessible at node {nid}")

        elif isinstance(item, _Timeout):
            res.errors.append(f"Request to node {nid} has timed out")

        elif isinstance(item, tuple):
            resp, exc = item
            res.value_per_node[nid] = resp
            res.errors.append(f"Assignment failed at node {nid}: {type(exc).__name__}: {exc}")

        else:
            res.value_per_node[nid] = item
            if item.value.empty and reg_val_str is not None:
                if optional_register:
                    res.warnings.append(f"Nonexistent register {reg_name!r} at node {nid} ignored as requested")
                else:
                    res.errors.append(f"Cannot assign nonexistent register {reg_name!r} at node {nid}")
    return res


class _NoService:
    pass


class _Timeout:
    pass


async def _access(
    local_node: pycyphal.application.Node,
    progress: Callable[[str], None],
    node_ids: Sequence[int],
    reg_name: str,
    reg_val_str: str | None,
    *,
    timeout: float,
) -> dict[
    int,
    Union[
        _NoService,
        _Timeout,
        "Access_1.Response",
        tuple["Access_1.Response", "pycyphal.application.register.ValueConversionError"],
    ],
]:
    from uavcan.register import Access_1

    out: dict[
        int,
        Access_1.Response | _NoService | _Timeout | pycyphal.application.register.ValueConversionError,
    ] = {}
    for nid in node_ids:
        progress(f"{nid: 5}: {reg_name!r}")
        cln = local_node.make_client(Access_1, nid)
        try:
            cln.response_timeout = timeout
            out[nid] = await _access_one(cln, reg_name, reg_val_str)
        finally:
            cln.close()
    return out


async def _access_one(
    client: pycyphal.presentation.Client["Access_1"],
    reg_name: str,
    reg_val_str: str | None,
) -> Union[
    _NoService,
    _Timeout,
    "Access_1.Response",
    tuple["Access_1.Response", "pycyphal.application.register.ValueConversionError"],
]:
    from uavcan.register import Access_1, Name_1
    from pycyphal.application.register import ValueProxy, ValueConversionError

    resp = await client(Access_1.Request(name=Name_1(reg_name)))
    if resp is None:
        return _NoService()
    assert isinstance(resp, Access_1.Response)
    if reg_val_str is None or resp.value.empty:  # Modification is not required or there is no such register.
        return resp

    # Coerce the supplied value to the type of the remote register.
    assert not resp.value.empty
    val = ValueProxy(resp.value)
    try:
        val.assign_environment_variable(reg_val_str)
    except ValueConversionError as ex:  # Oops, not coercible (e.g., register is float[], value is string)
        return resp, ex

    # Write the coerced value to the node; it may also modify it so return the response, not the coercion result.
    resp = await client(Access_1.Request(name=Name_1(reg_name), value=val.value))
    if resp is None:  # We got a response before but now we didn't, something is messed up so the result is different.
        return _Timeout()
    assert isinstance(resp, Access_1.Response)
    return resp


_logger = yakut.get_logger(__name__)



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Callable, Optional, Any, List, Dict
import inspect
import logging
import pycyphal
from pycyphal.transport import Transport as Transport
import click
from yakut.paths import OUTPUT_TRANSFER_ID_MAP_DIR, OUTPUT_TRANSFER_ID_MAP_MAX_AGE

TransportFactory = Callable[[], Optional[Transport]]
"""
The result is None if no transport configuration was provided when invoking the command.
"""

_logger = logging.getLogger(__name__)


def transport_factory_option(f: Callable[..., Any]) -> Callable[..., Any]:
    def validate(ctx: click.Context, param: object, value: Optional[str]) -> TransportFactory:
        from yakut import Purser

        _ = param
        _logger.debug("Transport expression: %r", value)

        def factory() -> Optional[Transport]:
            # Try constructing from the expression if provided:
            if value:
                try:
                    result = construct_transport(value)
                except Exception as ex:
                    raise click.BadParameter(f"Could not initialize transport {value!r}: {ex!r}") from ex
                _logger.info("Transport %r constructed from expression %r", result, value)
                return result
            # If no expression is given, construct from the registers passed via environment variables:
            try:
                from pycyphal.application import make_transport
            except (ImportError, AttributeError):
                _logger.info(
                    "Transport initialization expression is not provided and constructing the transport "
                    "from registers is not possible because the standard DSDL namespace is not compiled"
                )
                return None
            purser = ctx.find_object(Purser)
            assert isinstance(purser, Purser)
            result = make_transport(purser.get_registry())
            _logger.info("Transport constructed from registers: %r", result)
            return result

        return factory

    doc = f"""
Override the network interface configuration, including the local node-ID.
This option is only relevant for commands that access the network, like pub/sub/call/etc.; other commands ignore it.

By default, if this option is not given (neither via --transport nor YAKUT_TRANSPORT),
commands that access the network deduce the transport configuration from standard registers passed via
environment variables, such as UAVCAN__NODE__ID, UAVCAN__UDP__IFACE, UAVCAN__SERIAL__IFACE,
and so on.
The full list of registers that configure the transport is available in the definition of the standard
RPC-service "uavcan.register.Access", and in the documentation for PyCyphal:
https://pycyphal.readthedocs.io (see "make_transport()").
This method requires that the standard DSDL namespace "uavcan" is compiled (see command "yakut compile")!

If this expression is given, the registers are ignored, and the transport instance is constructed by evaluating it.
Upon evaluation, the expression should yield either a single transport instance or a sequence thereof.
In the latter case, the multiple transports will be joined under the same redundant transport instance,
which may be heterogeneous (e.g., UDP+Serial).
This method does not require any DSDL to be compiled at all.
To see supported transports and how they should be initialized, refer to https://pycyphal.readthedocs.io.

The expression does not need to explicitly reference the `pycyphal.transport` module
because its contents are wildcard-imported for convenience.
Further, when specifying a transport class, the suffix `Transport` may be omitted;
e.g., `UDPTransport` and `UDP` are equivalent.

Examples showcasing loopback, CAN, and heterogeneous UDP+Serial:

\b
    Loopback(None)
    CAN(can.media.socketcan.SocketCANMedia('vcan0',64),42)
    UDP('127.42.0.123',None),Serial("/dev/ttyUSB0",None)

To avoid issues caused by resetting the transfer-ID counters between invocations,
the tool stores the output transfer-ID map on disk keyed by the node-ID.
On this computer, the path is `{OUTPUT_TRANSFER_ID_MAP_DIR}`.
The map files can be removed to reset all transfer-ID counters to zero.
Files that are more than {OUTPUT_TRANSFER_ID_MAP_MAX_AGE} seconds old are not used.
"""
    f = click.option(
        "--transport",
        "-i",
        "transport_factory",
        envvar="YAKUT_TRANSPORT",
        type=str,
        metavar="EXPRESSION",
        callback=validate,
        help=doc,
    )(f)
    return f


def construct_transport(expression: str) -> Transport:
    context = _make_evaluation_context()
    trs = _evaluate_transport_expr(expression, context)
    _logger.debug("Transport expression evaluation result: %r", trs)
    if len(trs) == 1:
        return trs[0]  # Non-redundant transport
    if len(trs) > 1:
        from pycyphal.transport.redundant import RedundantTransport

        rt = RedundantTransport()
        for t in trs:
            rt.attach_inferior(t)
        assert rt.inferiors == trs
        return rt
    raise ValueError("No transports specified")


def _evaluate_transport_expr(expression: str, context: Dict[str, Any]) -> List[Transport]:
    out = eval(expression, context)
    if isinstance(out, Transport):
        return [out]
    if isinstance(out, (list, tuple)) and all(isinstance(x, Transport) for x in out):
        return list(out)
    raise ValueError(
        f"The expression {expression!r} yields an instance of {type(out).__name__!r}. "
        f"Expected an instance of pycyphal.transport.Transport or a list thereof."
    )


def _make_evaluation_context() -> Dict[str, Any]:
    import os

    def handle_import_error(parent_module_name: str, ex: ImportError) -> None:
        try:
            tr = parent_module_name.split(".")[2]
        except LookupError:
            tr = parent_module_name
        _logger.debug("Transport %r is not available due to the missing dependency %r", tr, ex.name)

    # This import is super slow, so we do it as late as possible.
    # Doing this when generating command-line arguments would be disastrous for performance.
    # noinspection PyTypeChecker
    pycyphal.util.import_submodules(pycyphal.transport, error_handler=handle_import_error)

    # Populate the context with all references that may be useful for the transport expression.
    context: Dict[str, Any] = {
        "pycyphal": pycyphal,
        "os": os,
    }

    # Expose pre-imported transport modules for convenience.
    for name, module in inspect.getmembers(pycyphal.transport, inspect.ismodule):
        if not name.startswith("_"):
            context[name] = module

    # Pre-import transport classes.
    for cls in pycyphal.util.iter_descendants(Transport):
        if not cls.__name__.startswith("_") and cls is not Transport:
            name = cls.__name__.rpartition(Transport.__name__)[0]
            assert name
            context[name] = cls
            context[cls.__name__] = cls

    _logger.debug("Transport expression evaluation context (on the next line):\n%r", context)
    return context



#!/usr/bin/env python3
#
# Copyright (C) 2020 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

import setuptools

setuptools.setup()



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import dataclasses
from typing import Sequence, TYPE_CHECKING, Callable
import bisect
import yakut

if TYPE_CHECKING:
    import pycyphal.application


@dataclasses.dataclass
class Result:
    names_per_node: dict[int, list[str] | None] = dataclasses.field(default_factory=dict)
    errors: list[str] = dataclasses.field(default_factory=list)
    warnings: list[str] = dataclasses.field(default_factory=list)


async def list_names(
    local_node: "pycyphal.application.Node",
    progress: Callable[[str], None],
    node_ids: Sequence[int],
    *,
    optional_service: bool,
    timeout: float,
) -> Result:
    res = Result()
    for nid, names in (await _impl_list_names(local_node, progress, node_ids, timeout=timeout)).items():
        _logger.debug("Names @%r: %r", nid, names)
        if isinstance(names, _NoService):
            res.names_per_node[nid] = None
            if optional_service:
                res.warnings.append(f"Register list service is not accessible at node {nid}, ignoring as requested")
            else:
                res.errors.append(f"Register list service is not accessible at node {nid}")
        else:
            lst = res.names_per_node.setdefault(nid, [])
            assert isinstance(lst, list)
            for idx, n in enumerate(names):
                if isinstance(n, _Timeout):
                    res.errors.append(f"Request #{idx} to node {nid} has timed out, data incomplete")
                else:
                    bisect.insort(lst, n)
    return res


class _NoService:
    pass


class _Timeout:
    pass


async def _impl_list_names(
    local_node: "pycyphal.application.Node",
    progress: Callable[[str], None],
    node_ids: Sequence[int],
    *,
    timeout: float,
) -> dict[int, list[str | _Timeout] | _NoService]:
    from uavcan.register import List_1

    out: dict[int, list[str | _Timeout] | _NoService] = {}
    for nid in node_ids:
        cln = local_node.make_client(List_1, nid)
        try:
            cln.response_timeout = timeout
            name_list: list[str | _Timeout] | _NoService = []
            for idx in range(2**16):
                progress(f"{nid: 5}: {idx: 5}")
                resp = await cln(List_1.Request(index=idx))
                assert isinstance(name_list, list)
                if resp is None:
                    if 0 == idx:  # First request timed out, assume service not supported or node is offline
                        name_list = _NoService()
                    else:  # Non-first request has timed out, assume network error
                        name_list.append(_Timeout())
                    break
                assert isinstance(resp, List_1.Response)
                name = resp.name.name.tobytes().decode(errors="replace")
                if not name:
                    break
                name_list.append(name)
        finally:
            cln.close()
        _logger.debug("Register names fetched from node %r: %r", nid, name_list)
        out[nid] = name_list
    return out


_logger = yakut.get_logger(__name__)



# Copyright (c) 2020 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

import sys
import pycyphal.util

# BY CONVENTION, the COMMAND and the MODULE it is defined in should be NAMED IDENTICALLY.
pycyphal.util.import_submodules(sys.modules[__name__])



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import logging
import sys
from typing import TYPE_CHECKING, TextIO, Union, Callable, Any
import click
import pycyphal
import yakut
from yakut.ui import ProgressReporter, show_error, show_warning
from yakut.param.formatter import FormatterHints
from yakut.yaml import Loader
from yakut.int_set_parser import parse_int_set, INT_SET_USER_DOC
from yakut.register import explode_value, get_access_response_metadata
from yakut.util import EXIT_CODE_UNSUCCESSFUL
from ._directive import Directive, SCHEMA_USER_DOC
from ._caller import do_calls, TypeCoercionFailure, Timeout, Skipped, Tag

if TYPE_CHECKING:
    import pycyphal.application
    from uavcan.register import Access_1

_logger = yakut.get_logger(__name__)


Predicate = Callable[["Access_1.Response"], bool]


_PREDICATES = {
    "m": lambda r: r if r.mutable else None,
    "i": lambda r: r if not r.mutable else None,
    "p": lambda r: r if r.persistent else None,
    "v": lambda r: r if not r.persistent else None,
    "mp": lambda r: r if r.mutable and r.persistent else None,
    "mv": lambda r: r if r.mutable and not r.persistent else None,
    "ip": lambda r: r if not r.mutable and r.persistent else None,
    "iv": lambda r: r if not r.mutable and not r.persistent else None,
}


_HELP = f"""
Read/write multiple registers at multiple nodes (useful for network configuration management).

Accepts a YAML/JSON file containing register names and/or values per node; the default is to read from stdin.
Acceptable formats are generated either by register-list (in which case registers will be only read)
or by this command (in which case the specified values will be written).

The registers will be processed strictly sequentially in the order they are specified
(this matters if register access has side effects).

Save configuration parameters into a file (using verbose form for clarity here):

\b
    yakut register-list 125     | yakut register-batch 125 --only=mp > pure_config.json
    yakut register-list 125,    | yakut register-batch     --only=mp > node_125_config.json
    yakut register-list 120-128 | yakut register-batch     --only=mp > network_config.json

To get human-friendly output either add --format=yaml to the last command, or pipe the output through "jq".

You can also remove the node-ID keys using jq '.[]' (this may be useful if you already have an existing file):

\b
    cat node_125_config.json | jq '.[]'

Apply the same parameters to nodes 10,11,12,13,14:

\b
    cat pure_config.json | y rb 10-15

You can also convert a pure config file that is not keyed by node-ID by adding the IDs using jq
(this is obviously not intended for interactive use):

\b
    cat pure_config.json | \\
    jq '. as $in | [range(10;15) | {{key: .|tostring, value: $in}}] | from_entries' | \\
    y rb

...same but applied to one node 125:

\b
    cat pure_config.json | jq '{{"125": .}}' | y rb

Filter output registers by name preserving the node-ID keys; in this case those matching "uavcan*id":

\b
    y rl 125, | y rb | jq 'map_values(with_entries(select(.key | test("uavcan.+id"))))'

Read diagnostic registers from two similar nodes
(protip: save the register names into a file instead of calling register-list each time):

\b
    y rl 124 | y rb 124,125 -oiv

{INT_SET_USER_DOC}
"""


@yakut.subcommand(aliases=["rbat", "rb"], help=_HELP)
@click.argument(
    "node_ids",
    required=False,
    type=lambda x: parse_int_set(x) if x is not None else None,
)
@click.option(
    "--file",
    "-f",
    type=click.File("r"),
    default=sys.stdin,
    help=f"""
Defaults to stdin. Supports YAML/JSON.

{SCHEMA_USER_DOC}
""",
)
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help="Service response timeout.",
)
@click.option(
    "--optional-register",
    "-r",
    is_flag=True,
    help="""
If modification is requested (i.e., if a value is given),
nodes that report that they don't have such register are silently ignored instead of reporting an error.
Best-effort output will always be produced regardless of this option; that is, it only affects the exit code.
""",
)
@click.option(
    "--detailed",
    "-d",
    count=True,
    help="Display the values as-is, with DSDL type information, do not simplify. Specify twice to include metadata.",
)
@click.option(
    "--only",
    "-o",
    type=click.Choice(list(_PREDICATES.keys()), case_sensitive=False),
    help="""
Filter the output to include only mutable/immutable, persistent/volatile registers.
All registers are always written regardless, this option only affects the final output.
""",
)
@yakut.pass_purser
@yakut.asynchronous()
async def register_batch(
    purser: yakut.Purser,
    node_ids: set[int] | int | None,
    file: TextIO,
    timeout: float,
    optional_register: bool,
    detailed: int,
    only: str | None,
) -> int:
    predicate: Predicate = _PREDICATES[only] if only else lambda _: True
    formatter = purser.make_formatter(FormatterHints(single_document=True))
    representer = _make_representer(detail=detailed)
    with file:
        directive = Directive.load(
            Loader().load(file),
            node_ids=node_ids if not isinstance(node_ids, int) else {node_ids},
        )
    _logger.debug("Loaded directive: %r", directive)
    with purser.get_node("register_batch", allow_anonymous=False) as node:
        with ProgressReporter() as prog:
            result = await do_calls(node, prog, directive=directive, timeout=timeout)
        if _logger.isEnabledFor(logging.INFO):
            _logger.info("%s", node.presentation.transport.sample_statistics())

    from uavcan.register import Access_1

    success = True

    def error(msg: str) -> None:
        nonlocal success
        success = False
        show_error(msg)

    for node_id, per_node in result.responses_per_node.items():
        failed_count = 0
        for reg_name, response in per_node.items():
            if isinstance(response, Access_1.Response) and not response.value.empty:
                continue
            failed_count += 1
            prefix = f"{node_id}:{reg_name!r}: "
            if isinstance(response, Access_1.Response) and response.value.empty:
                if optional_register:
                    show_warning(prefix + "No such register, ignored as requested")
                else:
                    error(prefix + "No such register")
            elif isinstance(response, TypeCoercionFailure):
                error(prefix + f"Original value unchanged because coercion failed: {response.msg}")
            elif isinstance(response, Timeout):
                error(prefix + "Timed out and gave up on this node")
            elif isinstance(response, Skipped):
                assert not success
            else:
                assert False, response
        _logger.info("Node %r: total %r, failed %r", node_id, len(per_node), failed_count)
        if failed_count > 0:
            show_warning(f"{node_id}: {failed_count} failed of {len(per_node)} total. Output incomplete.")

    final: Any = {
        node_id: {
            reg_name: representer(response)
            for reg_name, response in per_node.items()
            if isinstance(response, Access_1.Response) and predicate(response)
        }
        for node_id, per_node in result.responses_per_node.items()
    }
    if isinstance(node_ids, int):
        final = final[node_ids]
    sys.stdout.write(formatter(final))
    sys.stdout.flush()
    return 0 if success else EXIT_CODE_UNSUCCESSFUL


def _make_representer(detail: int) -> Callable[[Union[Tag, "Access_1.Response"]], Any]:
    from uavcan.register import Access_1

    def represent(response: Union[Tag, Access_1.Response]) -> Any:
        return (
            explode_value(
                response.value,
                simplify=detail < 1,
                metadata=get_access_response_metadata(response) if detail > 1 else None,
            )
            if isinstance(response, Access_1.Response)
            else None
        )

    return represent



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Any, Callable
import pycyphal
import yakut
from yakut.subject_resolver import SubjectResolver as SubjectResolver
from yakut import dtype_loader


class SubjectSpecifierProcessingError(RuntimeError):
    pass


class BadSpecifierError(SubjectSpecifierProcessingError):
    pass


class NoFixedPortIDError(SubjectSpecifierProcessingError):
    pass


class NetworkDiscoveryError(SubjectSpecifierProcessingError):
    pass


async def process_subject_specifier(
    specifier: str,
    resolver_provider: Callable[[], SubjectResolver],
) -> tuple[int, Any]:
    """
    Given the specifier string from the user return the subject-ID and the dtype for use with it.
    The resolver factory will be invoked if the user is asking us to perform automatic discovery through the network.
    If the specifier is complete (provides both subject-ID and dtype), the resolver will not be needed.
    This enables lazy construction of the local node and stuff, which is desirable as it may be very costly.
    The caller is responsible for closing the resolver afterwards (unless it was not needed).
    """
    if specifier.count(":") == 1:  # The simplest case -- full information is given explicitly.
        sp_sbj_id, sp_dty = specifier.split(":")
        _logger.info("Subject specifier interpreted as explicit: %r", (sp_sbj_id, sp_dty))
        subject_id = _parse_subject_id(sp_sbj_id)
        if not subject_id:
            raise BadSpecifierError(f"{subject_id} is not a valid subject-ID")
        return subject_id, dtype_loader.load_dtype(sp_dty)

    subject_id = _parse_subject_id(specifier)
    if subject_id is None:
        _logger.debug(
            "Subject specifier is not a valid subject-ID, assume it is dtype name with fixed port-ID: %r", specifier
        )
        dtype = dtype_loader.load_dtype(specifier)
        subject_id = pycyphal.dsdl.get_fixed_port_id(dtype)
        _logger.debug("Loaded dtype %s with fixed port-ID %r", dtype, subject_id)
        if subject_id is None:
            raise NoFixedPortIDError(
                f"Type specified as {specifier!r} is found but it has no fixed port-ID. "
                f"Consider specifying the subject-ID manually? The syntax is like 1234:{specifier}"
            )
        return subject_id, dtype

    _logger.info("Subject specifier is a number (%r), using network resolver (this may take a few seconds)", subject_id)
    assert isinstance(subject_id, int)
    resolver = resolver_provider()
    # Sorting to bubble newer types higher up. This should be natural sort.
    type_names = list(sorted(await resolver.dtypes_by_id(subject_id), reverse=True))
    _logger.debug("Dtype names found by the network resolver for subject-ID %s: %s", subject_id, type_names)
    try:
        dtype = next(
            dtype
            for dtype in (dtype_loader.load_dtype(tn, allow_minor_version_mismatch=True) for tn in type_names)
            if dtype is not None and pycyphal.dsdl.is_message_type(dtype) and not pycyphal.dsdl.is_service_type(dtype)
        )
    except StopIteration:
        raise NetworkDiscoveryError(
            f"Automatic network discovery did not return suitable dtypes for subject {subject_id}. "
            f"Either the subject-ID is incorrect, or the nodes that utilize it are currently offline, "
            f"or they do not support the introspection services required for automatic discovery. "
            f"Consider specifying the data type manually? The syntax is like {subject_id}:namespace.DataType"
        ) from None
    _logger.debug("Network discovery for subject %s done with dtype %s", subject_id, dtype)
    assert isinstance(dtype, type)
    return subject_id, dtype


def _parse_subject_id(spec: str) -> int | None:
    try:
        val = int(spec)
    except ValueError:
        return None
    if 0 <= val <= pycyphal.transport.MessageDataSpecifier.SUBJECT_ID_MASK:
        return val
    return None


_logger = yakut.get_logger(__name__)



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
from typing import TYPE_CHECKING, Callable, Sequence, Optional, Any
import asyncio
import click
import pycyphal
import yakut
from yakut.int_set_parser import parse_int_set, INT_SET_USER_DOC
from yakut.ui import ProgressReporter, show_error, show_warning
from yakut.param.formatter import FormatterHints
from yakut.util import EXIT_CODE_UNSUCCESSFUL

if TYPE_CHECKING:
    import pycyphal.application
    from uavcan.node import ExecuteCommand_1

_logger = yakut.get_logger(__name__)


_HELP = f"""
Invoke uavcan.node.ExecuteCommand on a group of nodes.

The response objects are returned as a per-node mapping unless the set of node-IDs is one integer,
in which case only that object is returned without the node-ID.

The command may be specified either as an explicit integer or as a mnemonic name defined in the
uavcan.node.ExecuteCommand service specification; e.g., restart=65535, begin_software_update=65533;
abbreviations are also accepted.

Restart many nodes at once, some of which may not be present (for Cyphal/CAN this would be the entire network):

\b
    yakut execute-command 0-128 restart -e
    yakut execute-command 0-128 65535 -e    # Same, with the command code given explicitly.

Request multiple nodes to install the same software image
(requires a file server node) (also see the file server command):

\b
    y cmd 122-126 begin_software_update /path/to/software/image

{INT_SET_USER_DOC}
"""


def _parse_status_set(inp: str) -> set[int] | None:
    """
    >>> _parse_status_set("") is None
    True
    >>> _parse_status_set("1") == {1}
    True
    >>> _parse_status_set("1-5") == {1, 2, 3, 4}
    True
    """
    if not inp:
        return None
    ins = parse_int_set(inp)
    return set(ins) if not isinstance(ins, (int, float)) else {ins}


@yakut.subcommand(aliases="cmd", help=_HELP)
@click.argument("node_ids", type=parse_int_set)
@click.argument("command")
@click.argument("parameter", default="")
@click.option(
    "--expect",
    "-e",
    type=_parse_status_set,
    default="0",
    metavar="[STATUS_CODES]",
    is_flag=False,
    flag_value="",
    help=f"""
Specify the set of status codes that shall be accepted as success (the default is only zero).
The set notation is supported; e.g., use 0-256 to accept any status code as success.

Using this option without any value (or an empty string)
will result in sending the requests simultaneously without requiring the nodes to respond and not validating the status
(fire and forget).
This is occasionally useful because some nodes may be unable to respond to certain requests such as COMMAND_RESTART;
some nodes may not support this service at all.

{INT_SET_USER_DOC}
""",
)
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help="Service response timeout.",
)
@yakut.pass_purser
@yakut.asynchronous()
async def execute_command(
    purser: yakut.Purser,
    node_ids: set[int] | int,
    command: str,
    parameter: str,
    expect: set[int] | None,
    timeout: float,
) -> int:
    _logger.debug(
        "node_ids=%r command=%r parameter=%r expect=%r timeout=%r", node_ids, command, parameter, expect, timeout
    )
    command_parsed = _parse_command(command)
    del command
    formatter = purser.make_formatter(FormatterHints(single_document=True))
    try:
        from uavcan.node import ExecuteCommand_1
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name)) from None

    request = ExecuteCommand_1.Request(command=command_parsed, parameter=parameter)
    # Ensure the parameters are valid before constructing the node.
    # Unfortunately, PyCyphal does not allow us to pass pre-serialized objects to a presentation-layer port instance.
    # Perhaps this capability should be added one day.
    _ = pycyphal.dsdl.serialize(request)

    with purser.get_node("execute_command", allow_anonymous=False) as node:
        with ProgressReporter() as prog:
            result = await _run(
                node,
                prog,
                list(sorted(node_ids)) if isinstance(node_ids, set) else [node_ids],
                request,
                timeout=timeout,
                fire_and_forget=expect is None,
            )

    success = True

    def error(text: str) -> None:
        nonlocal success
        success = False
        show_error(text)

    if expect is not None:
        for nid, resp in result.items():
            if resp is None:
                error(f"Timed out while waiting for response from node {nid}")
            elif resp.status not in expect:
                desc = _status_code_to_name(resp.status) or "(unknown status code)"
                error(f"Node {nid} returned unexpected status code: {resp.status} {desc}")
            else:
                _logger.debug("Success @%r: %r", nid, resp)
    else:
        show_warning("Responses not checked as requested")

    final: dict[int, Any] = {
        nid: pycyphal.dsdl.to_builtin(resp) if resp is not None else None for nid, resp in result.items()
    }
    if isinstance(node_ids, int):
        final = final[node_ids]
    sys.stdout.write(formatter(final))
    sys.stdout.flush()
    return 0 if success else EXIT_CODE_UNSUCCESSFUL


async def _run(
    local_node: "pycyphal.application.Node",
    progress: Callable[[str], None],
    node_ids: Sequence[int],
    request: "ExecuteCommand_1.Request",
    *,
    timeout: float,
    fire_and_forget: bool,
) -> dict[int, Optional["ExecuteCommand_1.Response"]]:
    from uavcan.node import ExecuteCommand_1

    async def once(nid: int) -> None:
        cln = local_node.make_client(ExecuteCommand_1, nid)
        try:
            cln.response_timeout = timeout
            result[nid] = await cln(request)
        finally:
            cln.close()

    result: dict[int, ExecuteCommand_1.Response | None] = {}
    if not fire_and_forget:
        for nid in node_ids:
            progress(f"{nid: 5}")
            await once(nid)
    else:
        for r in await asyncio.gather(
            *(asyncio.ensure_future(once(nid)) for nid in node_ids),
            return_exceptions=True,
        ):
            if isinstance(r, BaseException):
                raise r
        # noinspection PyTypeChecker
        result = dict(sorted(result.items()))

    return result


_DSDL_REQUEST_COMMAND_CONSTANT_PREFIX = "COMMAND_"
_DSDL_RESPONSE_STATUS_CONSTANT_PREFIX = "STATUS_"


def _parse_command(inp: str) -> int:
    """
    This function shall not be invoked before the command handler is entered because it requires the standard DSDL
    namespace to be available (or you would need to handle the DSDL not found error here).

    >>> _parse_command("0")
    0
    >>> _parse_command("0x100")
    256
    >>> _parse_command("restart") == _parse_command("ReStArT") == _parse_command("res") == 0xFFFF
    True
    >>> _parse_command("store_persistent_states") == 65530
    True
    >>> _parse_command("no matching option")  # doctest:+IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ClickException: ...
    >>> _parse_command("")  # doctest:+IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ClickException: ...
    """
    for fun in (int, lambda x: int(x, 0)):
        try:
            return fun(inp)  # type: ignore
        except ValueError:
            pass

    from uavcan.node import ExecuteCommand_1

    ty = ExecuteCommand_1.Request
    std = {
        x[len(_DSDL_REQUEST_COMMAND_CONSTANT_PREFIX) :]: getattr(ty, x)
        for x in dir(ty)
        if x.startswith(_DSDL_REQUEST_COMMAND_CONSTANT_PREFIX)
    }
    matches = {k: v for k, v in std.items() if k.lower().startswith(inp.lower())}
    if len(matches) == 1:
        return int(next(iter(matches.values())))
    raise click.ClickException(f"Command not understood: {inp!r}")


def _status_code_to_name(code: int) -> str | None:
    """
    >>> _status_code_to_name(0)
    'SUCCESS'
    >>> _status_code_to_name(1)
    'FAILURE'
    >>> _status_code_to_name(254) is None
    True
    """
    from uavcan.node import ExecuteCommand_1

    ty = ExecuteCommand_1.Response
    return {
        getattr(ty, x): x[len(_DSDL_RESPONSE_STATUS_CONSTANT_PREFIX) :]
        for x in dir(ty)
        if x.startswith(_DSDL_RESPONSE_STATUS_CONSTANT_PREFIX)
    }.get(code)
